import numpy as np
import math
import pandas as pd
import random


def estimated(N, p, m, x, w, b):
    s = np.zeros((N, m))
    y_est = np.zeros(N)
    for k in range(N):
        for i in range(m):
            for j in range(p):
                s[k, i] += x[k, j] * w[j, i]

    for k in range(N):
        s1 = 0
        for i in range(m):
            s1 += b[i] / (1 + math.exp(-s[k, i]))
        y_est[k] = s1
    return y_est


def error(N, y, y_est):
    E = 0
    for k in range(N):
        E += pow(y[k] - y_est[k], 2)
    return E


def derivatives(N, p, m, x, y, y_est, w, b):
    s = np.zeros((N, m))
    der_w = np.zeros((p, m))
    der_b = np.zeros(m)
    for k in range(N):
        for i in range(m):
            for j in range(p):
                s[k, i] += x[k, j] * w[j, i]
    for i in range(m):
        for j in range(p):
            sum1 = 0
            for k in range(N):
                r1 = y[k] - y_est[k]
                r2 = math.exp(-s[k, i]) / pow(1 + math.exp(-s[k, i]), 2)
                sum1 += r1 * r2 * x[k, j] * b[j]
            der_w[j, i] = (-2) * sum1
            sum2 = 0
            for k in range(N):
                sum2 += (y[k] - y_est[k]) / (1 + math.exp(-s[k, i]))
            der_b[i] = (-2) * sum2
    return der_w, der_b


# ---Feedforward Neural Network-------

N = 767
p = 8
m = 10
ni = 0.0001

df = pd.read_excel(r'x_Data.xlsx')
x = df.to_numpy()
df = pd.read_excel(r'y_Data.xlsx')
y = df.to_numpy()

#split train - test
N_train = round(0.6 * N)
N_test = N - N_train

x_train = np.zeros((N_train, p))
y_train = np.zeros(N_train)
for k in range(N_train):
    for j in range(p):
        x_train[k, j] = x[k, j]
    y_train[k] = y[k]

x_test = np.zeros((N_test, p))
y_test = np.zeros(N_test)
for k in range(N_test):
    for j in range(p):
        x_test[k, j] = x[N_train + k, j]
    y_test[k] = y[N_train + k]

#initialize weights
w = np.zeros((p, m))
for j in range(p):
    for i in range(m):
        w[j, i] = -0.1 + random.random() * 0.2

b = np.zeros(m)
for i in range(m):
    b[i] = -0.01 + random.random() * 0.02

#train
y_est_train = estimated(N_train, p, m, x_train, w, b)
E_train = error(N_train, y_train, y_est_train)
RMSE_train = math.sqrt(E_train / N_train)

y_est_test = estimated(N_test, p, m, x_test, w, b)
E_test = error(N_test, y_test, y_est_test)
RMSE_test = math.sqrt(E_test / N_test)

for it in range(50):
    der_w, der_b = derivatives(N_train, p, m, x_train, y_train, y_est_train, w, b)
    for j in range(p):
        for i in range(m):
            w[j, i] -= - ni * der_w[j, i]
    for i in range(m):
        b[i] -= ni * der_b[i]

    y_est_train = estimated(N_train, p, m, x_train, w, b)
    E_train = error(N_train, y_train, y_est_train)
    RMSE_train = math.sqrt(E_train / N_train)

    y_est_test = estimated(N_test, p, m, x_test, w, b)
    E_test = error(N_test, y_test, y_est_test)
    RMSE_test = math.sqrt(E_test / N_test)
    print(RMSE_train, RMSE_test)
